/**
 * Copyright (c) 2014-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */
#include <wdt/Wdt.h>
#include <wdt/test/TestCommon.h>

#include <gflags/gflags.h>
#include <glog/logging.h>
#include <gtest/gtest.h>

#include <folly/String.h>  // for humanify

using namespace std;

namespace facebook {
namespace wdt {

string genAllBytes() {
  string res;
  res.reserve(256);
  // put the \0 in the middle:
  for (int i = -128; i <= 127; ++i) {
    res.push_back(i);
  }
  LOG(INFO) << "Generated bytes : " << folly::humanify(res);
  return res;
}

TEST(RequestSerializationTest, UrlTests) {
  {
    WdtUri uri("wdt://blah.com?k1=v1&k2=v2&k3=v3.1,v3.2");
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "blah.com");
    EXPECT_EQ(uri.getQueryParam("k1"), "v1");
    EXPECT_EQ(uri.getQueryParam("k2"), "v2");
    EXPECT_EQ(uri.getQueryParam("k3"), "v3.1,v3.2");
    EXPECT_EQ(uri.getQueryParams().size(), 3);

    uri = "wdt://blah?name=first,second";
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getQueryParam("name"), "first,second");
    uri = "http://blah.com?name=test";
    ASSERT_TRUE(uri.getHostName().empty());

    uri = "wdt://localhost";
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "localhost");

    uri = "wdt://localhost.facebook.com?key=value1,value2";
    EXPECT_EQ(uri.getHostName(), "localhost.facebook.com");
    EXPECT_EQ(uri.getQueryParam("key"), "value1,value2");

    uri = "wdt://127.0.0.1?";
    EXPECT_EQ(uri.getHostName(), "127.0.0.1");
    EXPECT_EQ(uri.getQueryParams().size(), 0);

    // we ignore keys that don't have a value
    uri = "wdt://127.0.0.1?a";
    EXPECT_EQ(uri.getHostName(), "127.0.0.1");
    EXPECT_EQ(uri.getQueryParams().size(), 1);

    EXPECT_EQ(uri.generateUrl(), "wdt://127.0.0.1");

    uri = "wdt://?a=10";
    EXPECT_NE(uri.getErrorCode(), OK);

    WdtUri u1;
    u1.setQueryParam("a", "");  // will be skipped
    u1.setQueryParam("b", "bvalue");
    u1.setQueryParam("c", "");  // also skipped
    EXPECT_EQ(u1.generateUrl(), "wdt://?b=bvalue");

    vector<string> keys;
    vector<string> values;
    WdtUri wdtUri;
    for (int i = 0; i < 100; i++) {
      keys.push_back(to_string(rand32()));
      values.push_back(to_string(rand32()));
    }
    for (size_t i = 0; i < keys.size(); i++) {
      wdtUri.setQueryParam(keys[i], values[i]);
    }
    string urlStr = wdtUri.generateUrl();
    WLOG(INFO) << "Url with random numerical key/value " << urlStr;
    uri = urlStr;
    EXPECT_NE(uri.getErrorCode(), OK);
    ASSERT_TRUE(uri.getHostName().empty());
    for (size_t i = 0; i < keys.size(); i++) {
      EXPECT_EQ(uri.getQueryParam(keys[i]), values[i]);
    }
  }
  {
    // ipv6 uri test
    WdtUri uri("wdt://[::1]:22356?k1=v1");
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "::1");
    EXPECT_EQ(uri.getPort(), 22356);

    uri = "wdt://[::1]:";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "::1");
    EXPECT_EQ(uri.getPort(), -1);

    uri = "wdt://[12::12:1]:1";
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "12::12:1");
    EXPECT_EQ(uri.getPort(), 1);

    uri = "wdt://123.4.5.6:22356";
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "123.4.5.6");
    EXPECT_EQ(uri.getPort(), 22356);

    uri = "wdt://[";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "");

    uri = "wdt://[121";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "");

    uri = "wdt://[]";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "");

    uri = "wdt://[]:22356";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "");
    EXPECT_EQ(uri.getPort(), 22356);

    uri = "wdt://[::1]:*";
    EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
    EXPECT_EQ(uri.getHostName(), "::1");
    EXPECT_EQ(uri.getPort(), -1);

    uri = "wdt://]:22356";
    EXPECT_EQ(uri.getErrorCode(), OK);
    EXPECT_EQ(uri.getHostName(), "]");
    EXPECT_EQ(uri.getPort(), 22356);

    {
      uri = "wdt://[::1]:24689?num_ports=3";
      EXPECT_EQ(uri.getPort(), 24689);
      WdtTransferRequest transferRequest(uri.generateUrl());
      EXPECT_EQ(uri.getErrorCode(), OK);
      EXPECT_EQ(transferRequest.hostName, "::1");
      EXPECT_EQ(transferRequest.ports,
                WdtTransferRequest::genPortsVector(24689, 3));
    }
    {
      uri = "wdt://[::1]?num_ports=10";  // missing port
      WdtTransferRequest transferRequest(uri.generateUrl());
      EXPECT_EQ(uri.getErrorCode(), OK);
      EXPECT_EQ(transferRequest.errorCode, INVALID_REQUEST);
      EXPECT_EQ(transferRequest.hostName, "::1");
    }
    {
      uri = "wdt://[::1]:12345?num_ports=abc";  // bad num ports
      WdtTransferRequest transferRequest(uri.generateUrl());
      EXPECT_EQ(uri.getErrorCode(), OK);
      EXPECT_EQ(transferRequest.errorCode, URI_PARSE_ERROR);
      EXPECT_EQ(transferRequest.hostName, "::1");
    }
    {
      uri = "wdt://[::1]:24689?start_port=22356&ports=1,2,3,4";
      WdtTransferRequest transferRequest(uri.generateUrl());
      EXPECT_EQ(uri.getErrorCode(), OK);
      EXPECT_EQ(transferRequest.hostName, "::1");
      EXPECT_EQ(transferRequest.ports,
                WdtTransferRequest::genPortsVector(1, 4));
    }
    {
      // bad %sequence
      string ustr;
      ustr = "wdt://[::1]:24689?start_port=22356&ports=1,2,3,4&id=bar%xxblah";
      uri = ustr;
      WdtTransferRequest transferRequest(ustr);
      EXPECT_EQ(uri.getErrorCode(), URI_PARSE_ERROR);
      EXPECT_EQ(transferRequest.errorCode, URI_PARSE_ERROR);
      EXPECT_EQ(transferRequest.transferId, "bar");
    }
  }
  {
    WdtTransferRequest request(123, 5, "");
    request.hostName = "host1";
    const string binaryStr = genAllBytes();
    request.transferId = binaryStr;
    request.protocolVersion = 753;
    string serialized = request.genWdtUrlWithSecret();
    WLOG(INFO) << "Serialized " << serialized;
    WdtTransferRequest deser(serialized);
    EXPECT_EQ(deser.hostName, "host1");
    EXPECT_EQ(deser.transferId, binaryStr);
    EXPECT_EQ(deser.protocolVersion, 753);
    EXPECT_EQ(deser.ports.size(), 5);
    for (int i = 0; i < 5; ++i) {
      EXPECT_EQ(deser.ports[i], 123 + i);
    }
    EXPECT_EQ(deser.errorCode, OK);
    EXPECT_EQ(deser.genWdtUrlWithSecret(), serialized);
    EXPECT_EQ(deser, request);
  }
  {
    WdtTransferRequest transferRequest(24689, 1, "");
    // Lets not populate anything else
    transferRequest.hostName = "localhost";
    string serializedString = transferRequest.genWdtUrlWithSecret();
    WLOG(INFO) << serializedString;
    WdtTransferRequest dummy(serializedString);
    WLOG(INFO) << dummy.getLogSafeString();
    EXPECT_EQ(transferRequest, dummy);
  }
  {
    string uri = "wdt://localhost?ports=1&recpv=10";
    WdtTransferRequest transferRequest(uri);
    EXPECT_EQ(transferRequest.errorCode, OK);
    ASSERT_EQ(transferRequest.ports.size(), 1);
    EXPECT_EQ(transferRequest.ports[0], 1);
  }
  {
    WdtTransferRequest transferRequest(0, 8, "/dir3/dir4");
    Receiver receiver(transferRequest);
    transferRequest = receiver.init();
    EXPECT_EQ(transferRequest.errorCode, OK);
    ASSERT_TRUE(transferRequest.ports.size() != 0);
    for (auto port : transferRequest.ports) {
      ASSERT_TRUE(port != 0);
    }
    WLOG(INFO) << transferRequest.hostName;
    ASSERT_TRUE(!transferRequest.hostName.empty());
  }
  {
    string uri = "wdt://localhost?ports=1,2,3,10&dir=test&recpv=100&id=111";
    WdtTransferRequest transferRequest(uri);
    EXPECT_EQ(transferRequest.errorCode, OK);
    EXPECT_EQ(transferRequest.hostName, "localhost");
    EXPECT_EQ(transferRequest.directory, "test");
    EXPECT_EQ(transferRequest.protocolVersion, 100);
    EXPECT_EQ(transferRequest.transferId, "111");
    vector<int32_t> expectedPorts;
    for (int i = 0; i < 3; i++) {
      expectedPorts.push_back(i + 1);
    }
    expectedPorts.push_back(10);
    EXPECT_EQ(transferRequest.ports, expectedPorts);
  }
  {
    // missing numports/ports
    string uri = "wdt://localhost?dir=test&recpv=100&id=111";
    WdtTransferRequest transferRequest(uri);
    EXPECT_EQ(transferRequest.errorCode, INVALID_REQUEST);
    EXPECT_EQ(transferRequest.genWdtUrlWithSecret(), "INVALID_REQUEST");
  }
  {
    string url = "wdt://";
    WdtTransferRequest transferRequest(url);
    EXPECT_EQ(transferRequest.errorCode, INVALID_REQUEST);
    EXPECT_EQ(transferRequest.genWdtUrlWithSecret(), "INVALID_REQUEST");
  }
  {
    string url = "wdt://localhost:22355?num_ports=3";
    WdtTransferRequest transferRequest(url);
    Receiver receiver(transferRequest);
    auto retTransferRequest = receiver.init();
    EXPECT_EQ(retTransferRequest.errorCode, INVALID_REQUEST);
  }
  {
    string url = "wdt://localhost:22355";
    WdtTransferRequest transferRequest(url);
    Sender sender(transferRequest);
    auto retTransferRequest = sender.init();
    EXPECT_EQ(retTransferRequest.errorCode, INVALID_REQUEST);
  }
  {
    string url = "wdt://localhost:22355?num_ports=3";
    WdtTransferRequest transferRequest(url);
    EXPECT_EQ(transferRequest.errorCode, OK);
    transferRequest.directory = "blah";
    // Artificial error
    transferRequest.errorCode = ERROR;
    Receiver receiver(transferRequest);
    auto retTransferRequest = receiver.init();
    EXPECT_EQ(retTransferRequest.errorCode, ERROR);
  }
  {
    string uri = "wdt://localhost?dr=1";
    WdtTransferRequest transferRequest(uri);
    ASSERT_TRUE(transferRequest.downloadResumptionEnabled);
  }
  {
    string uri = "wdt://localhost?dr=0";
    WdtTransferRequest transferRequest(uri);
    ASSERT_TRUE(!transferRequest.downloadResumptionEnabled);
  }
  {
    string uri = "wdt://localhost?jimmy=1";
    WdtTransferRequest transferRequest(uri);
    ASSERT_TRUE(!transferRequest.downloadResumptionEnabled);
  }
  {
    for (int i = 0; i < 2; i++) {
      WdtTransferRequest transferRequest1(55000, 1, "");
      transferRequest1.hostName = "localhost";
      transferRequest1.downloadResumptionEnabled = (bool)i;
      WLOG(INFO) << "Download Resumption="
                 << transferRequest1.downloadResumptionEnabled;
      string serializedString = transferRequest1.genWdtUrlWithSecret();
      WLOG(INFO) << "serialized string=" << serializedString;
      WdtTransferRequest transferRequest(serializedString);
      Receiver receiver(transferRequest);
      transferRequest = receiver.init();
      EXPECT_EQ(transferRequest.downloadResumptionEnabled, (bool)i);
    }
  }
}

TEST(RequestSerializationTest, TransferIdGenerationTest) {
  string transferId1 = WdtBase::generateTransferId();
  string transferId2 = WdtBase::generateTransferId();
  EXPECT_NE(transferId1, transferId2);
}

TEST(TransferRequestTest, Encryption1) {
  {
    WdtTransferRequest req("");
    EXPECT_FALSE(req.encryptionData.isSet());
  }
  {
    WdtTransferRequest req(123, 3, "/foo/bar");
    WLOG(INFO) << "Url without encr= " << req.getLogSafeString();
    WdtTransferRequest req2(123, 3, "/foo/ba2");
    EXPECT_FALSE(req2 == req);
    req2.directory = "/foo/bar";
    EXPECT_EQ(req, req2);
    req.encryptionData = EncryptionParams(ENC_AES128_CTR, "data1");
    EXPECT_FALSE(req == req2);
  }
  {
    // TODO: hostname is mandatory in url yet not in constructor,
    // while directory isn't and yet is
    WdtTransferRequest req(123, 3, "");
    req.hostName = "host1";
    EXPECT_EQ(req.errorCode, OK);
    req.encryptionData = EncryptionParams(ENC_AES128_CTR, "barFOO");
    EXPECT_EQ(req.encryptionData.getType(), ENC_AES128_CTR);
    string binary("FOObar56");
    binary.push_back(0);
    binary.push_back(1);
    binary.push_back(0xff);
    binary.push_back(0xfe);
    const string secret(binary);
    req.encryptionData = EncryptionParams(ENC_AES128_CTR, secret);
    string ser = req.genWdtUrlWithSecret();
    WLOG(INFO) << "Url with e= " << ser;
    EXPECT_EQ(ser,
              "wdt://host1:123?enc=1:464f4f62617235360001fffe"
              "&num_ports=3&recpv=" +
                  std::to_string(Protocol::protocol_version));
    WdtTransferRequest unser(ser);
    EXPECT_EQ(unser.errorCode, OK);
    EXPECT_EQ(unser.encryptionData.getType(), ENC_AES128_CTR);
    EXPECT_EQ(unser.encryptionData.getSecret(), secret);
    EXPECT_EQ(req, unser);
  }
}

TEST(Wdt, WdtUnescape) {
  string inp("%41");
  string res;
  EXPECT_TRUE(WdtUri::unescape(res, folly::StringPiece(inp)));
  EXPECT_EQ(res, "A");
  inp = "B%43D";
  EXPECT_TRUE(WdtUri::unescape(res, folly::StringPiece(inp)));
  EXPECT_EQ(res, "ABCD");
  inp = "%45%a";
  EXPECT_FALSE(WdtUri::unescape(res, folly::StringPiece(inp)));
  EXPECT_EQ(res, "ABCDE");
  inp = "F%Xfstop1";
  EXPECT_FALSE(WdtUri::unescape(res, folly::StringPiece(inp)));
  inp = "G%fXstop2";
  EXPECT_FALSE(WdtUri::unescape(res, folly::StringPiece(inp)));
  inp = "H%";
  EXPECT_FALSE(WdtUri::unescape(res, folly::StringPiece(inp)));
  EXPECT_EQ(res, "ABCDEFGH");
}

TEST(Wdt, WdtInstanceTest) {
  Wdt& wdt1 = Wdt::initializeWdt("wdt");
  Wdt& wdt2 = Wdt::getWdt("wdt");
  Wdt& wdt3 = Wdt::getWdt("wdt");
  Wdt& wdt4 = Wdt::initializeWdt("wdt1");
  EXPECT_EQ(&wdt1, &wdt2);
  EXPECT_EQ(&wdt2, &wdt3);
  EXPECT_NE(&wdt1, &wdt4);
}
}
}  // namespace end

int main(int argc, char* argv[]) {
  FLAGS_logtostderr = true;
  testing::InitGoogleTest(&argc, argv);
  google::ParseCommandLineFlags(&argc, &argv, true);
  google::InitGoogleLogging(argv[0]);
  int ret = RUN_ALL_TESTS();
  return ret;
}
